# This module contains the "mkdocs-autorefs" plugin.
#
# After each page is processed by the Markdown converter, this plugin stores absolute URLs of every HTML anchors
# it finds to later be able to fix unresolved references.
#
# Once every page has been rendered and all identifiers and their URLs collected,
# the plugin fixes unresolved references in the HTML content of the pages.

from __future__ import annotations

import contextlib
import functools
import logging
from collections import defaultdict
from pathlib import PurePosixPath as URL  # noqa: N814
from typing import TYPE_CHECKING, Any, Callable, Literal
from urllib.parse import urlsplit
from warnings import warn

from mkdocs.config.base import Config
from mkdocs.config.config_options import Choice, Type
from mkdocs.plugins import BasePlugin, event_priority
from mkdocs.structure.pages import Page

from mkdocs_autorefs._internal.backlinks import Backlink, BacklinkCrumb
from mkdocs_autorefs._internal.references import AutorefsExtension, fix_refs, relative_url

if TYPE_CHECKING:
    from collections.abc import Sequence

    from jinja2.environment import Environment
    from mkdocs.config.defaults import MkDocsConfig
    from mkdocs.structure.files import Files
    from mkdocs.structure.nav import Section
    from mkdocs.structure.toc import AnchorLink

try:
    from mkdocs.plugins import get_plugin_logger

    _log = get_plugin_logger(__name__)
except ImportError:
    # TODO: Remove once support for MkDocs <1.5 is dropped.
    _log = logging.getLogger(f"mkdocs.plugins.{__name__}")  # type: ignore[assignment]


class AutorefsConfig(Config):
    """Configuration options for the `autorefs` plugin."""

    resolve_closest: bool = Type(bool, default=False)  # type: ignore[assignment]
    """Whether to resolve an autoref to the closest URL when multiple URLs are found for an identifier.

    By closest, we mean a combination of "relative to the current page" and "shortest distance from the current page".

    For example, if you link to identifier `hello` from page `foo/bar/`,
    and the identifier is found in `foo/`, `foo/baz/` and `foo/bar/baz/qux/` pages,
    autorefs will resolve to `foo/bar/baz/qux`, which is the only URL relative to `foo/bar/`.

    If multiple URLs are equally close, autorefs will resolve to the first of these equally close URLs.
    If autorefs cannot find any URL that is close to the current page, it will log a warning and resolve to the first URL found.

    When false and multiple URLs are found for an identifier, autorefs will log a warning and resolve to the first URL.
    """

    link_titles: bool | Literal["auto", "external"] = Choice((True, False, "auto", "external"), default="auto")  # type: ignore[assignment]
    """Whether to set titles on links.

    Such title attributes are displayed as tooltips when hovering over the links.

    - `"auto"`: autorefs will detect whether the instant preview feature of Material for MkDocs is enabled,
        and set titles on external links when it is, all links if it is not.
    - `"external"`: autorefs will set titles on external links only.
    - `True`: autorefs will set titles on all links.
    - `False`: autorefs will not set any title attributes on links.

    Titles are only set when they are different from the link's text.
    Titles are constructed from the linked heading's original title,
    optionally appending the identifier for API objects.
    """

    strip_title_tags: bool | Literal["auto"] = Choice((True, False, "auto"), default="auto")  # type: ignore[assignment]
    """Whether to strip HTML tags from link titles.

    Some themes support HTML in link titles, but others do not.

    - `"auto"`: strip tags unless the Material for MkDocs theme is detected.
    """


class AutorefsPlugin(BasePlugin[AutorefsConfig]):
    """The `autorefs` plugin for `mkdocs`.

    This plugin defines the following event hooks:

    - `on_config`, to configure itself
    - `on_page_markdown`, to set the current page in order for Markdown extension to use it
    - `on_env`, to apply cross-references once all pages have been rendered

    Check the [Developing Plugins](https://www.mkdocs.org/user-guide/plugins/#developing-plugins) page of `mkdocs`
    for more information about its plugin system.
    """

    scan_toc: bool = True
    """Whether to scan the table of contents for identifiers to map to URLs."""
    record_backlinks: bool = False
    """Whether to record backlinks."""
    current_page: Page | None = None
    """The current page being processed."""
    # YORE: Bump 2: Remove block.
    legacy_refs: bool = True
    """Whether to support legacy references."""

    def __init__(self) -> None:
        """Initialize the object."""
        super().__init__()

        # The plugin uses three URL maps, one for "primary" URLs, one for "secondary" URLs,
        # and one for "absolute" URLs.
        #
        # - A primary URL is an identifier that links to a specific anchor on a page.
        # - A secondary URL is an alias of an identifier that links to the same anchor as the identifier's primary URL.
        #   Primary URLs with these aliases as identifiers may or may not be rendered later.
        # - An absolute URL is an identifier that links to an external resource.
        #   These URLs are typically registered by mkdocstrings when loading object inventories.
        #
        # For example, mkdocstrings registers a primary URL for each heading rendered in a page.
        # Then, for each alias of this heading's identifier, it registers a secondary URL.
        #
        # We need to keep track of whether an identifier is primary or secondary,
        # to give it precedence when resolving cross-references.
        # We wouldn't want to log a warning if there is a single primary URL and one or more secondary URLs,
        # instead we want to use the primary URL without any warning.
        #
        # - A single primary URL mapped to an identifer? Use it.
        # - Multiple primary URLs mapped to an identifier? Use the first one, or closest one if configured as such.
        # - No primary URL mapped to an identifier, but a secondary URL mapped? Use it.
        # - Multiple secondary URLs mapped to an identifier? Use the first one, or closest one if configured as such.
        # - No secondary URL mapped to an identifier? Try using absolute URLs
        #   (typically registered by loading inventories in mkdocstrings).
        #
        # This logic unfolds in `_get_item_url`.
        self._primary_url_map: dict[str, list[str]] = {}
        self._secondary_url_map: dict[str, list[str]] = {}
        self._title_map: dict[str, str] = {}
        self._breadcrumbs_map: dict[str, BacklinkCrumb] = {}
        self._abs_url_map: dict[str, str] = {}
        self._backlinks: dict[str, dict[str, set[str]]] = defaultdict(lambda: defaultdict(set))
        # YORE: Bump 2: Remove line.
        self._get_fallback_anchor: Callable[[str], tuple[str, ...]] | None = None
        # YORE: Bump 2: Remove line.
        self._url_to_page: dict[str, Page] = {}

        self._link_titles: bool | Literal["external"] = True
        self._strip_title_tags: bool = False

    # ----------------------------------------------------------------------- #
    # MkDocs Hooks                                                            #
    # ----------------------------------------------------------------------- #
    def on_config(self, config: MkDocsConfig) -> MkDocsConfig | None:
        """Instantiate our Markdown extension.

        Hook for the [`on_config` event](https://www.mkdocs.org/user-guide/plugins/#on_config).
        In this hook, we instantiate our [`AutorefsExtension`][mkdocs_autorefs.AutorefsExtension]
        and add it to the list of Markdown extensions used by `mkdocs`.

        Arguments:
            config: The MkDocs config object.

        Returns:
            The modified config.
        """
        _log.debug("Adding AutorefsExtension to the list")
        config.markdown_extensions.append(AutorefsExtension(self))  # type: ignore[arg-type]

        # YORE: Bump 2: Remove block.
        # mkdocstrings still uses the `page` attribute as a string.
        # Fortunately, it does so in f-strings, so we can simply patch the `__str__` method
        # to render the URL.
        Page.__str__ = lambda page: page.url  # type: ignore[method-assign,attr-defined]

        if self.config.link_titles == "auto":
            if getattr(config.theme, "name", None) == "material" and "navigation.instant.preview" in config.theme.get(
                "features",
                (),
            ):
                self._link_titles = "external"
            else:
                self._link_titles = True
        else:
            self._link_titles = self.config.link_titles

        if self.config.strip_title_tags == "auto":
            if getattr(config.theme, "name", None) == "material" and "content.tooltips" in config.theme.get(
                "features",
                (),
            ):
                self._strip_title_tags = False
            else:
                self._strip_title_tags = True
        else:
            self._strip_title_tags = self.config.strip_title_tags

        return config

    def on_page_markdown(self, markdown: str, page: Page, **kwargs: Any) -> str:  # noqa: ARG002
        """Remember which page is the current one.

        Arguments:
            markdown: Input Markdown.
            page: The related MkDocs page instance.
            kwargs: Additional arguments passed by MkDocs.

        Returns:
            The same Markdown. We only use this hook to keep a reference to the current page URL,
                used during Markdown conversion by the anchor/heading scanner tree processors.
        """
        # YORE: Bump 2: Remove line.
        self._url_to_page[page.url] = page
        self.current_page = page
        return markdown

    def on_page_content(self, html: str, page: Page, **kwargs: Any) -> str:  # noqa: ARG002
        """Register breadcrumbs.

        Hook for the [`on_page_content` event](https://www.mkdocs.org/user-guide/plugins/#on_page_content).
        In this hook, we register breadcrumbs for each heading on each page.
        These breadcrumbs are used to provide contextual information in backlinks.

        Arguments:
            html: HTML converted from Markdown.
            page: The related MkDocs page instance.
            kwargs: Additional arguments passed by MkDocs.

        Returns:
            The same HTML.
        """
        self.current_page = page
        if self.record_backlinks:
            for item in page.toc.items:
                self._register_breadcrumbs(page, item)
        return html

    @event_priority(-50)  # Late, after mkdocstrings has finished loading inventories.
    def on_env(self, env: Environment, /, *, config: MkDocsConfig, files: Files) -> Environment:  # noqa: ARG002
        """Apply cross-references and collect backlinks.

        Hook for the [`on_env` event](https://www.mkdocs.org/user-guide/plugins/#on_env).
        In this hook, we try to fix unresolved references of the form `[title][identifier]` or `[identifier][]`.
        Doing that allows the user of `autorefs` to cross-reference objects in their documentation strings.
        It uses the native Markdown syntax so it's easy to remember and use.

        We log a warning for each reference that we couldn't map to an URL.

        We also collect backlinks at the same time. We fix cross-refs and collect backlinks in a single pass
        for performance reasons (we don't want to run the regular expression on each page twice).

        Arguments:
            env: The MkDocs environment.
            config: The MkDocs config object.
            files: The list of files in the MkDocs project.

        Returns:
            The unmodified environment.
        """
        for file in files:
            if file.page and file.page.content:
                _log.debug("Applying cross-refs in page %s", file.page.file.src_path)

                url_mapper = functools.partial(
                    self.get_item_url,
                    from_url=file.page.url,
                    # YORE: Bump 2: Remove line.
                    fallback=self.get_fallback_anchor,
                )
                backlink_recorder = (
                    functools.partial(self._record_backlink, page_url=file.page.url) if self.record_backlinks else None
                )
                # YORE: Bump 2: Replace `, _legacy_refs=self.legacy_refs` with `` within line.
                file.page.content, unmapped = fix_refs(
                    file.page.content,
                    url_mapper,
                    record_backlink=backlink_recorder,
                    link_titles=self._link_titles,
                    strip_title_tags=self._strip_title_tags,
                    _legacy_refs=self.legacy_refs,
                )

                if unmapped and _log.isEnabledFor(logging.WARNING):
                    for ref, context in unmapped:
                        message = f"from {context.filepath}:{context.lineno}: ({context.origin}) " if context else ""
                        _log.warning(
                            f"{file.page.file.src_path}: {message}Could not find cross-reference target '{ref}'",
                        )

        return env

    # ----------------------------------------------------------------------- #
    # Utilities                                                               #
    # ----------------------------------------------------------------------- #
    # YORE: Bump 2: Remove block.
    def map_urls(self, page: Page, anchor: AnchorLink) -> None:
        """Deprecated. Recurse on every anchor to map its ID to its absolute URL.

        This method is deprecated and will be removed in a future release.
        Please use the `register_anchor` or `register_url` methods instead.

        This method populates `self._primary_url_map` by side-effect.

        Arguments:
            page: The page containing the anchors.
            anchor: The anchor to process and to recurse on.
        """
        warn(
            "The `map_urls` method is deprecated and will be removed in a future release. "
            "Please use the `register_anchor` or `register_url` methods instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        self._map_urls(page, anchor)

    # YORE: Bump 2: Remove block.
    def _map_urls(self, page: Page | str, anchor: AnchorLink) -> None:
        if isinstance(page, str):
            try:
                page = self._url_to_page[page]
            except KeyError:
                page = self.current_page  # type: ignore[assignment]

        self.register_anchor(page, anchor.id, title=anchor.title, primary=True)  # type: ignore[arg-type]
        self._register_breadcrumbs(page, anchor)  # type: ignore[arg-type]

    def _register_breadcrumbs(self, page: Page, anchor: AnchorLink, parent: BacklinkCrumb | None = None) -> None:
        # Getting a breadcrumb has a side-effect of registering it in the breadcrumbs map.
        breadcrumb = self._get_breadcrumb(page, anchor, parent)
        for child in anchor.children:
            self._register_breadcrumbs(page, child, breadcrumb)

    def _get_breadcrumb(
        self,
        page: Page | Section,
        anchor: AnchorLink | None = None,
        parent: BacklinkCrumb | None = None,
    ) -> BacklinkCrumb:
        parent_breadcrumb = None if page.parent is None else self._get_breadcrumb(page.parent)
        if parent is None:
            if isinstance(page, Page):
                if (parent_url := page.url) not in self._breadcrumbs_map:
                    self._breadcrumbs_map[parent_url] = BacklinkCrumb(
                        title=page.title,
                        url=parent_url,
                        parent=parent_breadcrumb,
                    )
                parent = self._breadcrumbs_map[parent_url]
            else:
                parent = BacklinkCrumb(title=page.title, url="", parent=parent_breadcrumb)
        if anchor is None:
            return parent
        if (url := f"{page.url}#{anchor.id}") not in self._breadcrumbs_map:  # type: ignore[union-attr]
            # Skip the parent page if the anchor is a top-level heading, to reduce repetition.
            if anchor.level == 1:
                parent = parent.parent
            self._breadcrumbs_map[url] = BacklinkCrumb(title=anchor.title, url=url, parent=parent)
        return self._breadcrumbs_map[url]

    def _record_backlink(self, identifier: str, backlink_type: str, backlink_anchor: str, page_url: str) -> None:
        """Record a backlink.

        Arguments:
            identifier: The target identifier.
            backlink_type: The type of backlink.
            backlink_anchor: The backlink target anchor.
            page_url: The URL of the page containing the backlink.
        """
        # When we record backlinks, all identifiers have been registered.
        # If an identifier is not found in the primary or secondary URL maps, it's an absolute URL,
        # meaning it comes from an external source (typically an object inventory),
        # and we don't need to record backlinks for it.
        if identifier in self._primary_url_map or identifier in self._secondary_url_map:
            self._backlinks[identifier][backlink_type].add(f"{page_url}#{backlink_anchor}")

    def get_backlinks(self, *identifiers: str, from_url: str) -> dict[str, set[Backlink]]:
        """Return the backlinks to an identifier relative to the given URL.

        Arguments:
            *identifiers: The identifiers to get backlinks for.
            from_url: The URL of the page where backlinks are rendered.

        Returns:
            A dictionary of backlinks, with the type of reference as key and a set of backlinks as value.
            Each backlink is a tuple of (URL, title) tuples forming navigation breadcrumbs.
        """
        relative_backlinks: dict[str, set[Backlink]] = defaultdict(set)
        for identifier in set(identifiers):
            backlinks = self._backlinks.get(identifier, {})
            for backlink_type, backlink_urls in backlinks.items():
                for backlink_url in backlink_urls:
                    if backlink := self._get_backlink(from_url, backlink_url):
                        relative_backlinks[backlink_type].add(backlink)
        return relative_backlinks

    def _get_backlink(self, from_url: str, backlink_url: str) -> Backlink | None:
        breadcrumbs = []
        breadcrumb: BacklinkCrumb | None
        if not (breadcrumb := self._breadcrumbs_map.get(backlink_url)):
            _log.debug("No breadcrumb for backlink URL %s", backlink_url)
            return None
        while breadcrumb:
            breadcrumbs.append(
                BacklinkCrumb(
                    title=breadcrumb.title,
                    url=breadcrumb.url and relative_url(from_url, breadcrumb.url),
                    parent=breadcrumb.parent,
                ),
            )
            breadcrumb = breadcrumb.parent
        return Backlink(tuple(reversed(breadcrumbs)))

    def register_anchor(
        self,
        page: Page,
        identifier: str,
        anchor: str | None = None,
        *,
        title: str | None = None,
        primary: bool = True,
    ) -> None:
        """Register that an anchor corresponding to an identifier was encountered when rendering the page.

        Arguments:
            page: The page where the anchor was found.
            identifier: The identifier to register.
            anchor: The anchor on the page, without `#`. If not provided, defaults to the identifier.
            title: The title of the anchor (optional).
            primary: Whether this anchor is the primary one for the identifier.
        """
        # YORE: Bump 2: Remove block.
        if isinstance(page, str):
            try:
                page = self._url_to_page[page]
            except KeyError:
                page = self.current_page  # type: ignore[assignment]

        url = f"{page.url}#{anchor or identifier}"
        url_map = self._primary_url_map if primary else self._secondary_url_map
        if identifier in url_map:
            if url not in url_map[identifier]:
                url_map[identifier].append(url)
        else:
            url_map[identifier] = [url]
        if title and url not in self._title_map:
            self._title_map[url] = title

    def register_url(self, identifier: str, url: str) -> None:
        """Register that the identifier should be turned into a link to this URL.

        Arguments:
            identifier: The new identifier.
            url: The absolute URL (including anchor, if needed) where this item can be found.
        """
        self._abs_url_map[identifier] = url

    @staticmethod
    def _get_closest_url(from_url: str, urls: list[str], qualifier: str) -> str:
        """Return the closest URL to the current page.

        Arguments:
            from_url: The URL of the base page, from which we link towards the targeted pages.
            urls: A list of URLs to choose from.
            qualifier: The type of URLs we are choosing from.

        Returns:
            The closest URL to the current page.
        """
        base_url = URL(from_url)

        while True:
            if candidates := [url for url in urls if URL(url).is_relative_to(base_url)]:
                break
            base_url = base_url.parent
            if not base_url.name:
                break

        if not candidates:
            _log.warning(
                "Could not find closest %s URL (from %s, candidates: %s). "
                "Make sure to use unique headings, identifiers, or Markdown anchors (see our docs).",
                qualifier,
                from_url,
                urls,
            )
            return urls[0]

        winner = candidates[0] if len(candidates) == 1 else min(candidates, key=lambda c: c.count("/"))
        _log.debug("Closest URL found: %s (from %s, candidates: %s)", winner, from_url, urls)
        return winner

    def _get_urls(self, identifier: str) -> tuple[list[str], str]:
        try:
            return self._primary_url_map[identifier], "primary"
        except KeyError:
            return self._secondary_url_map[identifier], "secondary"

    def _get_item_url(
        self,
        identifier: str,
        from_url: str | None = None,
        # YORE: Bump 2: Remove line.
        fallback: Callable[[str], Sequence[str]] | None = None,
    ) -> str:
        try:
            urls, qualifier = self._get_urls(identifier)
        except KeyError:
            # YORE: Bump 2: Replace block with line 2.
            if identifier in self._abs_url_map:
                return self._abs_url_map[identifier]
            if fallback:
                new_identifiers = fallback(identifier)
                for new_identifier in new_identifiers:
                    with contextlib.suppress(KeyError):
                        url = self._get_item_url(new_identifier)
                        self._secondary_url_map[identifier] = [url]
                        return url
            raise

        if len(urls) > 1:
            if (self.config.resolve_closest or qualifier == "secondary") and from_url is not None:
                return self._get_closest_url(from_url, urls, qualifier)
            _log.warning(
                "Multiple %s URLs found for '%s': %s. "
                "Make sure to use unique headings, identifiers, or Markdown anchors (see our docs).",
                qualifier,
                identifier,
                urls,
            )
        return urls[0]

    def get_item_url(
        self,
        identifier: str,
        from_url: str | None = None,
        # YORE: Bump 2: Remove line.
        fallback: Callable[[str], Sequence[str]] | None = None,
    ) -> tuple[str, str | None]:
        """Return a site-relative URL with anchor to the identifier, if it's present anywhere.

        Arguments:
            identifier: The anchor (without '#').
            from_url: The URL of the base page, from which we link towards the targeted pages.

        Returns:
            A site-relative URL.
        """
        # YORE: Bump 2: Replace `, fallback` with `` within line.
        url = self._get_item_url(identifier, from_url, fallback)
        title = self._title_map.get(url) or None
        if from_url is not None:
            parsed = urlsplit(url)
            if not parsed.scheme and not parsed.netloc:
                url = relative_url(from_url, url)
        return url, title

    # YORE: Bump 2: Remove block.
    # ----------------------------------------------------------------------- #
    # Deprecated API                                                          #
    # ----------------------------------------------------------------------- #
    @property
    def get_fallback_anchor(self) -> Callable[[str], tuple[str, ...]] | None:
        """Fallback anchors getter."""
        return self._get_fallback_anchor

    # YORE: Bump 2: Remove block.
    @get_fallback_anchor.setter
    def get_fallback_anchor(self, value: Callable[[str], tuple[str, ...]] | None) -> None:
        """Fallback anchors setter."""
        self._get_fallback_anchor = value
        if value is not None:
            warn(
                "Setting a fallback anchor function is deprecated and will be removed in a future release.",
                DeprecationWarning,
                stacklevel=2,
            )



# Cross-references module.

from __future__ import annotations

import logging
import re
import warnings
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import lru_cache
from html import escape, unescape
from html.parser import HTMLParser
from io import StringIO
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Literal
from urllib.parse import urlsplit
from xml.etree.ElementTree import Element

from markdown.core import Markdown
from markdown.extensions import Extension
from markdown.extensions.toc import slugify
from markdown.inlinepatterns import REFERENCE_RE, ReferenceInlineProcessor
from markdown.treeprocessors import Treeprocessor
from markdown.util import HTML_PLACEHOLDER_RE, INLINE_PLACEHOLDER_RE
from markupsafe import Markup

from mkdocs_autorefs._internal.backlinks import BacklinksTreeProcessor

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path
    from re import Match

    from markdown import Markdown

    from mkdocs_autorefs._internal.plugin import AutorefsPlugin

try:
    from mkdocs.plugins import get_plugin_logger

    _log = get_plugin_logger(__name__)
except ImportError:
    # TODO: remove once support for MkDocs <1.5 is dropped
    _log = logging.getLogger(f"mkdocs.plugins.{__name__}")  # type: ignore[assignment]


AUTOREF_RE = re.compile(r"<autoref (?P<attrs>.*?)>(?P<title>.*?)</autoref>", flags=re.DOTALL)
"""The autoref HTML tag regular expression.

A regular expression to match mkdocs-autorefs' special reference markers
in the [`on_env` hook][mkdocs_autorefs.AutorefsPlugin.on_env].
"""


class AutorefsHookInterface(ABC):
    """An interface for hooking into how AutoRef handles inline references."""

    @dataclass
    class Context:
        """The context around an auto-reference."""

        domain: str
        """A domain like `py` or `js`."""
        role: str
        """A role like `class` or `function`."""
        origin: str
        """The origin of an autoref (an object identifier)."""
        filepath: str | Path
        """The path to the file containing the autoref."""
        lineno: int
        """The line number in the file containing the autoref."""

        def as_dict(self) -> dict[str, str]:
            """Convert the context to a dictionary of HTML attributes."""
            return {
                "domain": self.domain,
                "role": self.role,
                "origin": self.origin,
                "filepath": str(self.filepath),
                "lineno": str(self.lineno),
            }

    @abstractmethod
    def expand_identifier(self, identifier: str) -> str:
        """Expand an identifier in a given context.

        Parameters:
            identifier: The identifier to expand.

        Returns:
            The expanded identifier.
        """
        raise NotImplementedError

    @abstractmethod
    def get_context(self) -> AutorefsHookInterface.Context:
        """Get the current context.

        Returns:
            The current context.
        """
        raise NotImplementedError


class AutorefsInlineProcessor(ReferenceInlineProcessor):
    """A Markdown extension to handle inline references."""

    name: str = "mkdocs-autorefs"
    """The name of the inline processor."""

    hook: AutorefsHookInterface | None = None
    """The hook to use for expanding identifiers or adding context to autorefs."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(REFERENCE_RE, *args, **kwargs)

    # Code based on
    # https://github.com/Python-Markdown/markdown/blob/8e7528fa5c98bf4652deb13206d6e6241d61630b/markdown/inlinepatterns.py#L780
    def handleMatch(self, m: Match[str], data: str) -> tuple[Element | None, int | None, int | None]:  # type: ignore[override]  # noqa: N802
        """Handle an element that matched.

        Arguments:
            m: The match object.
            data: The matched data.

        Returns:
            A new element or a tuple.
        """
        text, index, handled = self.getText(data, m.end(0))
        if not handled:
            return None, None, None

        identifier, slug, end, handled = self._eval_id(data, index, text)
        if not handled or identifier is None:
            return None, None, None

        if slug is None and re.search(r"[\x00-\x1f]", identifier):
            # Do nothing if the matched reference still contains control characters (from 0 to 31 included)
            # that weren't unstashed when trying to compute a slug of the title.
            return None, m.start(0), end

        return self._make_tag(identifier, text, slug=slug), m.start(0), end

    def _unstash(self, identifier: str) -> str:
        stashed_nodes: dict[str, Element | str] = self.md.treeprocessors["inline"].stashed_nodes  # type: ignore[attr-defined]

        def _repl(match: Match) -> str:
            el = stashed_nodes.get(match[1])
            if isinstance(el, Element):
                return f"`{''.join(el.itertext())}`"
            if el == "\x0296\x03":
                return "`"
            return str(el)

        return INLINE_PLACEHOLDER_RE.sub(_repl, identifier)

    def _eval_id(self, data: str, index: int, text: str) -> tuple[str | None, str | None, int, bool]:
        """Evaluate the id portion of `[ref][id]`.

        If `[ref][]` use `[ref]`.

        Arguments:
            data: The data to evaluate.
            index: The starting position.
            text: The text to use when no identifier.

        Returns:
            A tuple containing the identifier, its optional slug, its end position, and whether it matched.
        """
        m = self.RE_LINK.match(data, pos=index)
        if not m:
            return None, None, index, False

        if identifier := m.group(1):
            # An identifier was provided, match it exactly (later).
            slug = None
        else:
            # Only a title was provided, use it as identifier.
            identifier = text

            # Catch single stash entries, like the result of [`Foo`][].
            if match := INLINE_PLACEHOLDER_RE.fullmatch(identifier):
                stashed_nodes: dict[str, Element | str] = self.md.treeprocessors["inline"].stashed_nodes  # type: ignore[attr-defined]
                el = stashed_nodes.get(match[1])
                if isinstance(el, Element) and el.tag == "code":
                    # The title was wrapped in backticks, we only keep the content,
                    # and tell autorefs to match the identifier exactly.
                    identifier = "".join(el.itertext())
                    slug = None
                    # Special case: allow pymdownx.inlinehilite raw <code> snippets but strip them back to unhighlighted.
                    if match := HTML_PLACEHOLDER_RE.fullmatch(identifier):
                        stash_index = int(match.group(1))
                        html = self.md.htmlStash.rawHtmlBlocks[stash_index]
                        identifier = Markup(html).striptags()  # noqa: S704
                        self.md.htmlStash.rawHtmlBlocks[stash_index] = escape(identifier)

            # In any other case, unstash the title and slugify it.
            # Examples: ``[`Foo` and `Bar`]``, `[The *Foo*][]`.
            else:
                identifier = self._unstash(identifier)
                slug = slugify(identifier, separator="-")

        end = m.end(0)
        return identifier, slug, end, True

    def _make_tag(self, identifier: str, text: str, *, slug: str | None = None) -> Element:
        """Create a tag that can be matched by `AUTO_REF_RE`.

        Arguments:
            identifier: The identifier to use in the HTML property.
            text: The text to use in the HTML tag.

        Returns:
            A new element.
        """
        el = Element("autoref")
        if self.hook:
            identifier = self.hook.expand_identifier(identifier)
            el.attrib.update(self.hook.get_context().as_dict())
        el.set("identifier", identifier)
        el.text = text
        if slug:
            el.attrib["slug"] = slug
        return el


class AnchorScannerTreeProcessor(Treeprocessor):
    """Tree processor to scan and register HTML anchors."""

    name: str = "mkdocs-autorefs-anchors-scanner"
    """The name of the tree processor."""

    _htags: ClassVar[set[str]] = {"h1", "h2", "h3", "h4", "h5", "h6"}

    def __init__(self, plugin: AutorefsPlugin, md: Markdown | None = None) -> None:
        """Initialize the tree processor.

        Parameters:
            plugin: A reference to the autorefs plugin, to use its `register_anchor` method.
        """
        super().__init__(md)
        self._plugin = plugin

    def run(self, root: Element) -> None:
        """Run the tree processor.

        Arguments:
            root: The root element of the tree.
        """
        if self._plugin.current_page is not None:
            pending_anchors = _PendingAnchors(self._plugin)
            self._scan_anchors(root, pending_anchors)
            pending_anchors.flush()

    def _scan_anchors(self, parent: Element, pending_anchors: _PendingAnchors, last_heading: str | None = None) -> None:
        for el in parent:
            if el.tag == "a":
                # We found an anchor. Record its id if it has one.
                if anchor_id := el.get("id"):
                    pending_anchors.append(anchor_id)
                # If the element has text or a link, it's not an alias.
                # Non-whitespace text after the element interrupts the chain, aliases can't apply.
                if el.text or el.get("href") or (el.tail and el.tail.strip()):
                    pending_anchors.flush(title=last_heading)

            elif el.tag == "p":
                # A `p` tag is a no-op for our purposes, just recurse into it in the context
                # of the current collection of anchors.
                self._scan_anchors(el, pending_anchors, last_heading)
                # Non-whitespace text after the element interrupts the chain, aliases can't apply.
                if el.tail and el.tail.strip():
                    pending_anchors.flush()

            elif el.tag in self._htags:
                # If the element is a heading, that turns the pending anchors into aliases.
                last_heading = el.text
                pending_anchors.flush(el.get("id"), title=last_heading)

            else:
                # But if it's some other interruption, flush anchors anyway as non-aliases.
                pending_anchors.flush(title=last_heading)
                # Recurse into sub-elements, in a *separate* context.
                self.run(el)


class HeadingScannerTreeProcessor(Treeprocessor):
    """Tree processor to scan and register HTML headings."""

    name: str = "mkdocs-autorefs-headings-scanner"
    """The name of the tree processor."""

    _htags: ClassVar[set[str]] = {"h1", "h2", "h3", "h4", "h5", "h6"}

    def __init__(self, plugin: AutorefsPlugin, md: Markdown | None = None) -> None:
        """Initialize the tree processor.

        Parameters:
            plugin: A reference to the autorefs plugin, to use its `register_anchor` method.
        """
        super().__init__(md)
        self._plugin = plugin

    def run(self, root: Element) -> None:
        """Run the tree processor.

        Arguments:
            root: The root element of the tree.
        """
        if self._plugin.current_page is not None:
            self._scan_headings(root)

    def _scan_headings(self, parent: Element) -> None:
        for el in parent:
            if el.tag in self._htags:
                if h_id := el.get("id"):
                    self._plugin.register_anchor(
                        self._plugin.current_page,  # type: ignore[arg-type]
                        h_id,
                        title=el.text,
                    )
            else:
                self._scan_headings(el)


class AutorefsExtension(Extension):
    """Markdown extension that transforms unresolved references into auto-references.

    Auto-references are then resolved later by the MkDocs plugin.

    This extension also scans Markdown anchors (`[](){#some-id}`)
    to register them with the MkDocs plugin.
    """

    def __init__(
        self,
        plugin: AutorefsPlugin | None = None,
        **kwargs: Any,
    ) -> None:
        """Initialize the Markdown extension.

        Parameters:
            plugin: An optional reference to the autorefs plugin (to pass it to the anchor scanner tree processor).
            **kwargs: Keyword arguments passed to the [base constructor][markdown.Extension].
        """
        super().__init__(**kwargs)
        self.plugin = plugin
        """A reference to the autorefs plugin."""

    def extendMarkdown(self, md: Markdown) -> None:  # noqa: N802 (casing: parent method's name)
        """Register the extension.

        Add an instance of our [`AutorefsInlineProcessor`][mkdocs_autorefs.AutorefsInlineProcessor] to the Markdown parser.
        Also optionally add an instance of our [`AnchorScannerTreeProcessor`][mkdocs_autorefs.AnchorScannerTreeProcessor]
        and [`BacklinksTreeProcessor`][mkdocs_autorefs.BacklinksTreeProcessor] to the Markdown parser
        if a reference to the autorefs plugin was passed to this extension.

        Arguments:
            md: A `markdown.Markdown` instance.
        """
        md.inlinePatterns.register(
            AutorefsInlineProcessor(md),
            AutorefsInlineProcessor.name,
            priority=168,  # Right after markdown.inlinepatterns.ReferenceInlineProcessor
        )
        if self.plugin is not None:
            # Scan headings to register them.
            if self.plugin.scan_toc:
                md.treeprocessors.register(
                    HeadingScannerTreeProcessor(self.plugin, md),
                    HeadingScannerTreeProcessor.name,
                    priority=0,
                )
            # Markdown anchors require the `attr_list` extension.
            if self.plugin.scan_toc and "attr_list" in md.treeprocessors:
                _log_enabling_markdown_anchors()
                md.treeprocessors.register(
                    AnchorScannerTreeProcessor(self.plugin, md),
                    AnchorScannerTreeProcessor.name,
                    priority=0,
                )
            # Backlinks require IDs on headings, which are either set by `toc`,
            # or manually by the user with `attr_list`.
            if self.plugin.record_backlinks and ("attr_list" in md.treeprocessors or "toc" in md.treeprocessors):
                _log_enabling_backlinks()
                md.treeprocessors.register(
                    BacklinksTreeProcessor(self.plugin, md),
                    BacklinksTreeProcessor.name,
                    priority=0,
                )


class _PendingAnchors:
    """A collection of HTML anchors that may or may not become aliased to an upcoming heading."""

    def __init__(self, plugin: AutorefsPlugin):
        self.plugin = plugin
        self.anchors: list[str] = []

    def append(self, anchor: str) -> None:
        self.anchors.append(anchor)

    def flush(self, alias_to: str | None = None, title: str | None = None) -> None:
        if page := self.plugin.current_page:
            for anchor in self.anchors:
                self.plugin.register_anchor(page, anchor, alias_to, title=title, primary=True)
            self.anchors.clear()


class _AutorefsAttrs(dict):
    _handled_attrs: ClassVar[set[str]] = {
        "identifier",
        "optional",
        "hover",  # TODO: Remove at some point.
        "class",
        "domain",
        "role",
        "origin",
        "filepath",
        "lineno",
        "slug",
        "backlink-type",
        "backlink-anchor",
    }

    @property
    def context(self) -> AutorefsHookInterface.Context | None:
        try:
            return AutorefsHookInterface.Context(
                domain=self["domain"],
                role=self["role"],
                origin=self["origin"],
                filepath=self["filepath"],
                lineno=int(self["lineno"]),
            )
        except KeyError:
            return None

    @property
    def remaining(self) -> str:
        return " ".join(k if v is None else f'{k}="{v}"' for k, v in self.items() if k not in self._handled_attrs)


class _HTMLAttrsParser(HTMLParser):
    def __init__(self):
        super().__init__()
        self.attrs = {}

    def parse(self, html: str) -> _AutorefsAttrs:
        self.reset()
        self.attrs.clear()
        self.feed(html)
        return _AutorefsAttrs(self.attrs)

    def handle_starttag(self, tag: str, attrs: list[tuple[str, str | None]]) -> None:  # noqa: ARG002
        self.attrs.update(attrs)


class _HTMLTagStripper(HTMLParser):
    def __init__(self) -> None:
        super().__init__()
        self.text = StringIO()

    def strip(self, html: str) -> str:
        self.reset()
        self.text = StringIO()
        self.feed(html)
        return self.text.getvalue()

    def handle_data(self, data: str) -> None:
        self.text.write(data)


def relative_url(url_a: str, url_b: str) -> str:
    """Compute the relative path from URL A to URL B.

    Arguments:
        url_a: URL A.
        url_b: URL B.

    Returns:
        The relative URL to go from A to B.
    """
    parts_a = url_a.split("/")
    url_b, *rest = url_b.split("#", 1)
    anchor = rest[0] if rest else ""
    parts_b = url_b.split("/")

    # Remove common left parts.
    while parts_a and parts_b and parts_a[0] == parts_b[0]:
        parts_a.pop(0)
        parts_b.pop(0)

    # Go up as many times as remaining a parts' depth.
    levels = len(parts_a) - 1
    parts_relative = [".."] * levels + parts_b
    relative = "/".join(parts_relative)
    return f"{relative}#{anchor}"


def fix_ref(
    url_mapper: Callable[[str], tuple[str, str | None]],
    unmapped: list[tuple[str, AutorefsHookInterface.Context | None]],
    record_backlink: Callable[[str, str, str], None] | None = None,
    *,
    link_titles: bool | Literal["external"] = True,
    strip_title_tags: bool = False,
) -> Callable:
    """Return a `repl` function for [`re.sub`](https://docs.python.org/3/library/re.html#re.sub).

    In our context, we match Markdown references and replace them with HTML links.

    When the matched reference's identifier was not mapped to an URL, we append the identifier to the outer
    `unmapped` list. It generally means the user is trying to cross-reference an object that was not collected
    and rendered, making it impossible to link to it. We catch this exception in the caller to issue a warning.

    Arguments:
        url_mapper: A callable that gets an object's site URL by its identifier,
            such as [mkdocs_autorefs.AutorefsPlugin.get_item_url][].
        unmapped: A list to store unmapped identifiers.
        record_backlink: A callable to record backlinks.
        link_titles: How to set HTML titles on links. Always (`True`), never (`False`), or external-only (`"external"`).
        strip_title_tags: Whether to strip HTML tags from link titles.

    Returns:
        The actual function accepting a [`Match` object](https://docs.python.org/3/library/re.html#match-objects)
        and returning the replacement strings.
    """

    def inner(match: Match) -> str:
        title = match["title"]
        attrs = _html_attrs_parser.parse(f"<a {match['attrs']}>")
        identifier: str = attrs["identifier"]
        slug = attrs.get("slug", None)
        optional = "optional" in attrs

        identifiers = (identifier, slug) if slug else (identifier,)

        if (
            record_backlink
            and (backlink_type := attrs.get("backlink-type"))
            and (backlink_anchor := attrs.get("backlink-anchor"))
        ):
            record_backlink(identifier, backlink_type, backlink_anchor)

        try:
            url, original_title = _find_url(identifiers, url_mapper)
        except KeyError:
            if optional:
                _log.debug("Unresolved optional cross-reference: %s", identifier)
                return f'<span title="{identifier}">{title}</span>'
            unmapped.append((identifier, attrs.context))
            if title == identifier:
                return f"[{identifier}][]"
            if title == f"<code>{identifier}</code>" and not slug:
                return f"[<code>{identifier}</code>][]"
            return f"[{title}][{identifier}]"

        parsed = urlsplit(url)
        external = parsed.scheme or parsed.netloc

        classes = (attrs.get("class") or "").strip().split()
        classes = ["autorefs", "autorefs-external" if external else "autorefs-internal", *classes]
        class_attr = " ".join(classes)

        if remaining := attrs.remaining:
            remaining = f" {remaining}"

        title_attr = ""
        if link_titles is True or (link_titles == "external" and external):
            if optional:
                # The `optional` attribute is generally only added by mkdocstrings handlers,
                # for API objects, meaning we can and should append the full identifier.
                tooltip = _tooltip(identifier, original_title, strip_tags=strip_title_tags)
            else:
                # Autorefs without `optional` are generally user-written ones,
                # so we should only use the original title.
                tooltip = original_title or ""

            if tooltip and tooltip not in f"<code>{title}</code>":
                title_attr = f' title="{_html_tag_stripper.strip(tooltip) if strip_title_tags else escape(tooltip)}"'

        return f'<a class="{class_attr}"{title_attr} href="{escape(url)}"{remaining}>{title}</a>'

    return inner


def fix_refs(
    html: str,
    url_mapper: Callable[[str], tuple[str, str | None]],
    *,
    record_backlink: Callable[[str, str, str], None] | None = None,
    link_titles: bool | Literal["external"] = True,
    strip_title_tags: bool = False,
    # YORE: Bump 2: Remove line.
    _legacy_refs: bool = True,
) -> tuple[str, list[tuple[str, AutorefsHookInterface.Context | None]]]:
    """Fix all references in the given HTML text.

    Arguments:
        html: The text to fix.
        url_mapper: A callable that gets an object's site URL by its identifier,
            such as [mkdocs_autorefs.AutorefsPlugin.get_item_url][].
        record_backlink: A callable to record backlinks.
        link_titles: How to set HTML titles on links. Always (`True`), never (`False`), or external-only (`"external"`).
        strip_title_tags: Whether to strip HTML tags from link titles.

    Returns:
        The fixed HTML, and a list of unmapped identifiers (string and optional context).
    """
    unmapped: list[tuple[str, AutorefsHookInterface.Context | None]] = []
    html = AUTOREF_RE.sub(
        fix_ref(url_mapper, unmapped, record_backlink, link_titles=link_titles, strip_title_tags=strip_title_tags),
        html,
    )

    # YORE: Bump 2: Remove block.
    if _legacy_refs:
        html = AUTO_REF_RE.sub(_legacy_fix_ref(url_mapper, unmapped), html)

    return html, unmapped


def _find_url(
    identifiers: Iterable[str],
    url_mapper: Callable[[str], tuple[str, str | None]],
) -> tuple[str, str | None]:
    for identifier in identifiers:
        try:
            return url_mapper(identifier)
        except KeyError:
            pass
    raise KeyError(f"None of the identifiers {identifiers} were found")


def _tooltip(identifier: str, title: str | None, *, strip_tags: bool = False) -> str:
    if title:
        # Don't append identifier if it's already in the title.
        if identifier in title:
            return title
        # Append identifier (useful for API objects).
        if strip_tags:
            return f"{title} ({identifier})"
        return f"{title} (<code>{identifier}</code>)"
    # No title, just return the identifier.
    if strip_tags:
        return identifier
    return f"<code>{identifier}</code>"


@lru_cache
def _log_enabling_markdown_anchors() -> None:
    _log.debug("Enabling Markdown anchors feature")


@lru_cache
def _log_enabling_backlinks() -> None:
    _log.debug("Enabling backlinks feature")


_html_attrs_parser = _HTMLAttrsParser()
_html_tag_stripper = _HTMLTagStripper()


# YORE: Bump 2: Remove block.
_ATTR_VALUE = r'"[^"<>]+"|[^"<> ]+'  # Possibly with double quotes around
AUTO_REF_RE = re.compile(
    rf"<span data-(?P<kind>autorefs-(?:identifier|optional|optional-hover))=(?P<identifier>{_ATTR_VALUE})"
    rf"(?: class=(?P<class>{_ATTR_VALUE}))?(?P<attrs> [^<>]+)?>(?P<title>.*?)</span>",
    flags=re.DOTALL,
)
"""Deprecated. Use [`AUTOREF_RE`][mkdocs_autorefs.AUTOREF_RE] instead."""


# YORE: Bump 2: Remove block.
def __getattr__(name: str) -> Any:
    if name == "AutoRefInlineProcessor":
        warnings.warn("AutoRefInlineProcessor was renamed AutorefsInlineProcessor", DeprecationWarning, stacklevel=2)
        return AutorefsInlineProcessor
    raise AttributeError(f"module 'mkdocs_autorefs.references' has no attribute {name}")


# YORE: Bump 2: Remove block.
def _legacy_fix_ref(
    url_mapper: Callable[[str], tuple[str, str | None]],
    unmapped: list[tuple[str, AutorefsHookInterface.Context | None]],
) -> Callable:
    """Return a `repl` function for [`re.sub`](https://docs.python.org/3/library/re.html#re.sub).

    In our context, we match Markdown references and replace them with HTML links.

    When the matched reference's identifier was not mapped to an URL, we append the identifier to the outer
    `unmapped` list. It generally means the user is trying to cross-reference an object that was not collected
    and rendered, making it impossible to link to it. We catch this exception in the caller to issue a warning.

    Arguments:
        url_mapper: A callable that gets an object's site URL by its identifier,
            such as [mkdocs_autorefs.AutorefsPlugin.get_item_url][].
        unmapped: A list to store unmapped identifiers.

    Returns:
        The actual function accepting a [`Match` object](https://docs.python.org/3/library/re.html#match-objects)
        and returning the replacement strings.
    """

    def inner(match: Match) -> str:
        identifier = match["identifier"].strip('"')
        title = match["title"]
        kind = match["kind"]
        attrs = match["attrs"] or ""
        classes = (match["class"] or "").strip('"').split()

        try:
            url, _ = url_mapper(unescape(identifier))
        except KeyError:
            if kind == "autorefs-optional":
                return title
            if kind == "autorefs-optional-hover":
                return f'<span title="{identifier}">{title}</span>'
            unmapped.append((identifier, None))
            if title == identifier:
                return f"[{identifier}][]"
            return f"[{title}][{identifier}]"

        warnings.warn(
            "autorefs `span` elements are deprecated in favor of `autoref` elements: "
            f'`<span data-autorefs-identifier="{identifier}">...</span>` becomes `<autoref identifer="{identifier}">...</autoref>`',
            DeprecationWarning,
            stacklevel=1,
        )
        parsed = urlsplit(url)
        external = parsed.scheme or parsed.netloc
        classes = ["autorefs", "autorefs-external" if external else "autorefs-internal", *classes]
        class_attr = " ".join(classes)
        if kind == "autorefs-optional-hover":
            return f'<a class="{class_attr}" title="{identifier}" href="{escape(url)}"{attrs}>{title}</a>'
        return f'<a class="{class_attr}" href="{escape(url)}"{attrs}>{title}</a>'

    return inner



"""mkdocs-autorefs package.

Automatically link across pages in MkDocs.
"""

from __future__ import annotations

from mkdocs_autorefs._internal.backlinks import Backlink, BacklinkCrumb, BacklinksTreeProcessor
from mkdocs_autorefs._internal.plugin import AutorefsConfig, AutorefsPlugin
from mkdocs_autorefs._internal.references import (
    AUTO_REF_RE,
    AUTOREF_RE,
    AnchorScannerTreeProcessor,
    AutorefsExtension,
    AutorefsHookInterface,
    AutorefsInlineProcessor,
    HeadingScannerTreeProcessor,
    fix_ref,
    fix_refs,
    relative_url,
)

__all__: list[str] = [
    "AUTOREF_RE",
    "AUTO_REF_RE",
    "AnchorScannerTreeProcessor",
    "AutorefsConfig",
    "AutorefsExtension",
    "AutorefsHookInterface",
    "AutorefsInlineProcessor",
    "AutorefsPlugin",
    "Backlink",
    "BacklinkCrumb",
    "BacklinksTreeProcessor",
    "HeadingScannerTreeProcessor",
    "fix_ref",
    "fix_refs",
    "relative_url",
]



"""Deprecated. Import from 'mkdocs_autorefs' instead."""

# YORE: Bump 2: Remove file.

import warnings
from typing import Any

from mkdocs_autorefs._internal import references


def __getattr__(name: str) -> Any:
    warnings.warn(
        "Importing from 'mkdocs_autorefs.references' is deprecated. Import directly from 'mkdocs_autorefs' instead.",
        DeprecationWarning,
        stacklevel=2,
    )
    return getattr(references, name)



# Backlinks module.

from __future__ import annotations

import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, ClassVar

from markdown.core import Markdown
from markdown.treeprocessors import Treeprocessor

if TYPE_CHECKING:
    from xml.etree.ElementTree import Element

    from markdown import Markdown

    from mkdocs_autorefs._internal.plugin import AutorefsPlugin

try:
    from mkdocs.plugins import get_plugin_logger

    _log = get_plugin_logger(__name__)
except ImportError:
    # TODO: remove once support for MkDocs <1.5 is dropped
    _log = logging.getLogger(f"mkdocs.plugins.{__name__}")  # type: ignore[assignment]


@dataclass(frozen=True, order=True)
class BacklinkCrumb:  # noqa: PLW1641
    """A navigation breadcrumb for a backlink."""

    title: str
    """The title of the breadcrumb."""
    url: str
    """The URL of the breadcrumb."""
    parent: BacklinkCrumb | None = None
    """The parent breadcrumb."""

    def __eq__(self, value: object) -> bool:
        """Compare URLs for equality."""
        if isinstance(value, BacklinkCrumb):
            return self.url == value.url
        return False


@dataclass(eq=True, frozen=True, order=True)
class Backlink:
    """A backlink (list of breadcrumbs)."""

    crumbs: tuple[BacklinkCrumb, ...]
    """The list of breadcrumbs."""


class BacklinksTreeProcessor(Treeprocessor):
    """Enhance autorefs with `backlink-type` and `backlink-anchor` attributes.

    These attributes are then used later to register backlinks.
    """

    name: str = "mkdocs-autorefs-backlinks"
    """The name of the tree processor."""
    initial_id: str | None = None
    """The initial heading ID."""

    _htags: ClassVar[set[str]] = {"h1", "h2", "h3", "h4", "h5", "h6"}

    def __init__(self, plugin: AutorefsPlugin, md: Markdown | None = None) -> None:
        """Initialize the tree processor.

        Parameters:
            plugin: A reference to the autorefs plugin, to use its `register_anchor` method.
        """
        super().__init__(md)
        self._plugin = plugin
        self._last_heading_id: str | None = None

    def run(self, root: Element) -> None:
        """Run the tree processor.

        Arguments:
            root: The root element of the document.
        """
        if self._plugin.current_page is not None:
            self._last_heading_id = self.initial_id
            self._enhance_autorefs(root)

    def _enhance_autorefs(self, parent: Element) -> None:
        for el in parent:
            if el.tag in self._htags:
                self._last_heading_id = el.get("id")
            elif el.tag == "autoref":
                if "backlink-type" not in el.attrib:
                    el.set("backlink-type", "referenced-by")
                if "backlink-anchor" not in el.attrib and self._last_heading_id:
                    el.set("backlink-anchor", self._last_heading_id)
            else:
                self._enhance_autorefs(el)
